/*  File Version: 4.0.3
 *	Copyright, Chris Ruddell, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the
 *	following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list
 *	of conditions and the following disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors
 *	may be used to endorse or promote products derived from this software without specific
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 *	OF SUCH DAMAGE.
 */
package com.buzztouch.core

import android.content.Context
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.R
import com.buzztouch.data_models.BT_item
import com.buzztouch.data_models.BtAppConfig
import com.buzztouch.data_models.BtConfig
import com.buzztouch.data_models.db_models.BtRootItem
import com.buzztouch.database.BT_repository
import com.buzztouch.extensions.*
import com.buzztouch.utils.*
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import org.json.JSONObject
import java.io.IOException
import kotlin.coroutines.CoroutineContext

/*
		Application Configuration File / Data
		---------------------------------------------------------------------------
		One file holds all the configuration data associated with the application. This file must exist
		in the res/raw folder in the project. This file is normally named bt_config.txt and can be
		read / edited with a normal text editor. If this configuration data uses a dataURL,
		a remote server will be polled for content changes. Changes will be downloaded and
		saved locally. Once	this happens, the bt_config.txt file is no longer used and instead the application
		refers to it's newly downloaded and cached data.

		In other words, if a dataURL is used then the
		configuration file in the Eclipse project is only referenced so it can find the buzztouchAppId, buzztouchAPIKey,
		and dataURL. After that, it uses the data that was saved from the URL.
		If no dataURL is provided, the file in the bundle will be read and parsed everytime the app is started.

		Example 1 - "offline app"
		First launch, app reads from BT_config.txt in res/raw directory, parses and saves to database

		Subsequent launches, app reads from database and loads config, then reads from res/raw directory.
		If the version number of the res/raw config data has changed (can happen with an app update),
		app parses res/raw and saves to database.

		Example 2 - "online app"
		First launch, app reads from BT_config.txt in the res/raw directory, parses and saves to database.
		App then queries the reportToCloudUrl to see if lastModifiedDate is different than res/raw date.
		If so, parse the downloaded data and save to database.

		Subsequent launches, app reads from database and loads config, then reads from res/raw directory.
		If the version number of the res/raw config data has changed (can happen with app update),
		app parses res/raw version and saves to database.

		After both the database and the res/raw config have been parsed, app queries the reportToCloudUrl
		to get the lastModifiedDate.  If it is different than the cached version, app downloads fresh
		data from the dataURL, parses and saves to database.  Note that the query of the cloud will happen
		asynchronously, controlled by the BT_activity_host's onResume method.
	 */
object BT_configuration : CoroutineScope {
    override val coroutineContext: CoroutineContext get() = Dispatchers.IO
    private var dataURL = ""
    private var reportToCloudURL = ""
    private val offlineApp: Boolean get() = buzztouch_appDelegate.rootApp.dataURL.isEmpty()

    val canReportToCloud: Boolean get() = dataURL.isNotEmpty() && reportToCloudURL.isNotEmpty()

    // loadAppConfigData is helper...
    suspend fun loadAppConfigData(): Boolean {
        // update device connection type...
        buzztouch_appDelegate.rootApp.rootDevice?.updateDeviceConnectionType()

        // update device size...
        buzztouch_appDelegate.rootApp.rootDevice?.updateDeviceSize()

        // process config data using local/cached copy
        return processConfigData()
    }

    private suspend fun getLocalConfig(): BtAppConfig? {
        BtLog("getLocalConfig()")
        val bundledConfigJson: String? = getBundledConfig(appContext)
        val bundledConfig: BtAppConfig? = bundledConfigJson?.fromJson<BtConfig>()?.BT_appConfig
        val cachedConfig: BtAppConfig? = loadFromDatabase()

        if (cachedConfig == null) {
            BtLog("Unable to load config from db - using bundled config instead")
            BT_repository.saveAppConfig(bundledConfigJson ?: "") // save to database
            return loadFromDatabase() // try to load back from database so we get items with jsonVars
        }

        val bundledVersion: Float = bundledConfig?.BT_items?.firstOrNull()?.version?.toFloatOrNull() ?: 0.0f
        val cachedVersion: Float = cachedConfig.BT_items.firstOrNull()?.version?.toFloatOrNull() ?: bundledVersion
        BtLog("getLocalConfig - using ${if (bundledVersion > cachedVersion) "bundled config" else "cached config"} because bundledVersion:$bundledVersion and cachedVersion:$cachedVersion")

        return if (bundledVersion > cachedVersion) {
            BT_repository.saveAppConfig(bundledConfigJson ?: "")
            loadFromDatabase()
        }
        else cachedConfig
    }

    private suspend fun loadFromDatabase(): BtAppConfig? = BT_repository.getAppConfig()?.loadValues()

    private fun BtAppConfig?.isOnlineApp(): Boolean = this?.BT_items?.firstOrNull()?.let { it.reportToCloudURL.isSet() && it.dataURL.isSet() } ?: false

    private fun BtAppConfig.loadValues(): BtAppConfig? {
        val rootItem = BT_items.firstOrNull() ?: return null
        buzztouch_appDelegate.rootApp.apply {
            themes = rootItem.BT_themes?.toArrayList() ?: ArrayList()
            tabs = rootItem.BT_tabs?.toArrayList() ?: ArrayList()
            menus = rootItem.BT_menus?.toArrayList() ?: ArrayList()
            screens = rootItem.BT_screens?.toArrayList() ?: ArrayList()
            itemId = rootItem.itemId ?: ""
            itemType = rootItem.itemType ?: ""
            buzztouchAppId = rootItem.buzztouchAppId ?: ""
            buzztouchAPIKey = rootItem.buzztouchAPIKey ?: ""
            dataURL = rootItem.dataURL ?: ""
            reportToCloudURL = rootItem.reportToCloudURL ?: ""
            registerForPushURL = rootItem.registerForPushURL ?: ""
            name = rootItem.name ?: ""
            rootItem.version?.toFloat()?.let { setVersion(it) }
            currentMode = rootItem.currentMode ?: ""
            startLocationUpdates = rootItem.startLocationUpdates == "1"
            promptForPushNotifications = rootItem.promptForPushNotifications ?: ""
            allowRotation = rootItem.allowRotation ?: ""
            homeScreenItemId = rootItem.homeScreenItemId ?: homeScreenItemId
        }

        return this
    }

    private suspend fun processCachedData(context: Context): Boolean {
        BtLog("processConfigData()")

        if (!BT_fileManager.doesCachedFileExist(CACHED_CONFIGURATION_NAME)) return true

        val cachedData = BT_fileManager.readTextFileFromCache(CACHED_CONFIGURATION_NAME)


        return true
    }

    private suspend fun processConfigData(): Boolean {
        BtLog("processConfigData()")

        // check to see if cached copy of bt_config.txt exists.  If so, use that.  Otherwise use the one from the res/raw directory
        val localConfig: BtAppConfig = getLocalConfig() ?: return false
        
        val rootItem: BtRootItem = localConfig.BT_items.firstOrNull() ?: return false
        
        reportToCloudURL = rootItem.reportToCloudURL ?: ""
        dataURL = rootItem.dataURL ?: ""

        val dataWasProcessed: Boolean = withContext(Dispatchers.IO) { parseAppJSONData(BtConfig(localConfig)) }

        // check for error
        if (!dataWasProcessed) return false

        return true
    }

    fun refreshData(listener: BT_ConfigurationListener) {
        BtLog("$TAG: refreshData(listener)")
        launch {
            BT_preferencesKt.setDataBoostrapDone(false)
            val objDownloader = BT_downloader(dataURL)
            objDownloader.saveAsFileName = ""
            val downloadedData: String? = dataURL.readTextFromUrl("").getIfSet()
            val config: BtAppConfig? = downloadedData?.let { BT_repository.saveAppConfig(it) }
            val dataWasProcessed = config?.let { parseAppJSONData(BtConfig(it)) } ?: false
            
            if (dataWasProcessed) listener.onConfigurationProcessed()
            else listener.onErrorProcessingConfiguration()
        }
    }

    // checks if data needs to be refreshed.
    fun dataIsDirty(listener: BT_ConfigurationListener) {
        BtLog("$TAG: dataIsDirty(listener)")
        launch {
            val dataIsDirty = dataIsDirty()
            listener.onDirtyCheck(dataIsDirty)
        }
    }

    // private boolean check if updated config data exists    
    private suspend fun dataIsDirty(): Boolean {
        var dataIsDirty = false

        // must have a dataUrl to download config data
        if (dataURL.isEmpty() || reportToCloudURL.isEmpty()) {
            BT_debugger.logError(TAG, "dataIsDirty() -- ERROR: NO URL TO CHECK:" + if (dataURL.isEmpty()) "dataURL is empty" else "reportToCloudURL is empty")
            return false
        }

        // first, check if we have a dataUrl set but do not have a cached copy available
        if (!BT_fileManager.doesCachedFileExist(CACHED_CONFIGURATION_NAME)) {
            BtLog("$TAG: dataIsDirty check - no cached config file exists!")
            dataIsDirty = true
        } else {
            val lastModified = BT_fileManagerKt.readTextFileFromCache(LAST_MODIFIED_NAME)
            BtLog("$TAG: dataIsDirty check - lastModified = $lastModified")
            val url = "$reportToCloudURL&currentMode=${buzztouch_appDelegate.rootApp.currentMode}"
            val resObj = withContext(Dispatchers.IO) { readTextFromUrl(url) }
            val status = resObj.first
            val body = resObj.second
            if (status < 300) {
                BtLog("$TAG: dataIsDirty check - lastModified server version:$body")

                // save downloaded lastModifiedDate
                BT_fileManager.saveTextFileToCache(body, LAST_MODIFIED_NAME)
                if (lastModified != body) {
                    BtLog("dataIsDirty true because cached: $lastModified != server($status): $body")
                    dataIsDirty = true
                }
            } else {
                BT_debugger.logError(TAG, "dataIsDirty ERROR! ($status): $body")
                dataIsDirty = true
            }
        }
        BtLog("$TAG: dataIsDirty - " + if (dataIsDirty) "true" else "false")
        return dataIsDirty
    }

    private fun parseAppJSONData(obj: BtConfig): Boolean {
        BtLog("parseAppJSONData()")
        BT_debugger.showIt("$TAG: parseJSONData")

        // init properties
        dataURL = ""
        
        val rootItem: BtRootItem = obj.BT_appConfig.BT_items.firstOrNull() ?: return false
        // set the string properties for this app
        BtLog("$TAG: parsing core settings...")
        BtLog("$TAG : ${rootItem.BT_menus}")
        obj.BT_appConfig.loadValues()
        dataURL = rootItem.dataURL ?: dataURL
        reportToCloudURL = rootItem.reportToCloudURL ?: reportToCloudURL

        BT_preferencesKt.setDataBoostrapDone(true)
        BT_preferencesKt.setConfigVersion(buzztouch_appDelegate.rootApp.configVersion)

        BT_debugger.showIt("$TAG:parseJSONData done parsing application data")

        // send it back..
        return true
    }

    interface BT_ConfigurationListener {
        fun onConfigurationProcessed()
        fun onErrorProcessingConfiguration()
        fun onDirtyCheck(dirty: Boolean)
    }

    private const val TAG = "BT_configuration"
    const val CACHED_CONFIGURATION_NAME = "cachedAppConfig.txt"
    const val LAST_MODIFIED_NAME = "appModified.txt"

    fun getBundledConfig(context: Context?): String? {
        return try {
            BT_fileManager.parseResource(context, R.raw.bt_config)
        } catch (e: IOException) {
            e.printStackTrace()
            null
        }
    }
}
