package com.buzztouch.database

import androidx.annotation.WorkerThread
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.core.BT_configuration
import com.buzztouch.data_models.BT_item
import com.buzztouch.data_models.BtAppConfig
import com.buzztouch.data_models.BtConfig
import com.buzztouch.data_models.db_models.BtScreen
import com.buzztouch.extensions.*
import com.buzztouch.utils.BT_fileManager
import com.buzztouch.utils.BT_strings
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import org.json.JSONArray
import org.json.JSONException
import org.json.JSONObject

object BT_repository {
    const val KEY_ITEM_ID = "itemId"
    const val KEY_ITEM_NICKNAME = "itemNickname"
    const val KEY_ITEM_TYPE = "itemType"
    const val KEY_CHILD_ITEMS = "childItems"

    @WorkerThread
    fun getScreenById(itemId: String): BT_item? = buzztouch_appDelegate.theContext?.let { BT_roomDatabase.getScreenById(it, itemId) }

    suspend fun getAppConfig(): BtAppConfig? {
        val config = withContext(Dispatchers.IO) { BT_roomDatabase[appContext].configDao().getConfig() } ?: return null
        var rootItem = config.BT_items.firstOrNull() ?: return null
        val items = withContext(Dispatchers.IO) {
            BT_roomDatabase[appContext].dao().getAll().let { allItems ->
                val childItems = allItems.filter { it.type == BT_item.Type.ChildItem }
                allItems.map { it.initFromDb(childItems) }
            }
        }
        rootItem = rootItem
            .copy(BT_themes = items.filter { it.type == BT_item.Type.Theme })
            .copy(BT_menus = items.filter { it.type == BT_item.Type.Menu })
            .copy(BT_tabs = items.filter { it.type == BT_item.Type.Tab })
            .copy(BT_screens = items.filter { it.type == BT_item.Type.Screen })

        return if (rootItem.BT_screens.isNullOrEmpty()) null else config.copy(BT_items = listOf(rootItem))
    }

    private fun JSONObject.findItemJson(key: String, itemId: String): JSONObject? {
        return this
            .getJsonObjectOrNull("BT_appConfig")
            ?.getJsonArrayOrNull("BT_items")
            ?.getJsonObjectOrNull(0)
            ?.let {
                it.getJsonArrayOrNull(key)?.listObjects()?.firstOrNull { it.getStringOrNull("itemId") == itemId }
            }
    }

    private fun JSONObject.findChildItemById(key: String, parentItemId: String, childItemId: String): JSONObject? {
        return this
            .findItemJson(key, parentItemId)
            ?.getJsonArrayOrNull("childItems")
            ?.listObjects()
            ?.firstOrNull { it.getStringOrNull("itemId") == childItemId }
    }

    suspend fun saveAppConfig(configJson: String): BtAppConfig {
        val config: BtAppConfig
        withContext(Dispatchers.IO) {
            // first, save the config data
            val jsonObject = JSONObject(configJson)
            val configObj: BtConfig = configJson.fromJson<BtConfig>() ?: BtConfig(BtAppConfig(emptyList()))
            config = configObj.BT_appConfig
            BT_roomDatabase[appContext].configDao().clearAll()
            BT_roomDatabase[appContext].configDao().saveConfig(config)

            // second, save the BT_items
            BT_roomDatabase[appContext].dao().clear()
            config.BT_items.flatMap { it.BT_themes ?: emptyList() }.forEachIndexed { index, item ->
                insertBtItemIntoDatabase(item, index, null, BT_item.Type.Theme, jsonObject.findItemJson("BT_themes", item.itemId))
            }
            config.BT_items.flatMap { it.BT_menus ?: emptyList() }.forEachIndexed { index, item ->
                insertBtItemIntoDatabase(item, index, null, BT_item.Type.Menu, jsonObject.findItemJson("BT_menus", item.itemId))
                item.childItems?.forEachIndexed { i, childItem ->
                    insertBtItemIntoDatabase(childItem, i, item.itemId, BT_item.Type.ChildItem, jsonObject.findChildItemById("BT_menus", item.itemId, childItem.itemId))
                }
            }
            config.BT_items.flatMap { it.BT_tabs ?: emptyList() }.forEachIndexed { index, item ->
                insertBtItemIntoDatabase(item, index, null, BT_item.Type.Tab, jsonObject.findItemJson("BT_tabs", item.itemId))
                item.childItems?.forEachIndexed { i, childItem ->
                    insertBtItemIntoDatabase(childItem, i, item.itemId, BT_item.Type.ChildItem, jsonObject.findChildItemById("BT_tabs", item.itemId, childItem.itemId))
                }
            }
            config.BT_items.flatMap { it.BT_screens ?: emptyList() }.forEachIndexed { index, item ->
                insertBtItemIntoDatabase(item, index, null, BT_item.Type.Screen, jsonObject.findItemJson("BT_screens", item.itemId))
                item.childItems?.forEachIndexed { i, childItem ->
                    insertBtItemIntoDatabase(childItem, i, item.itemId, BT_item.Type.ChildItem, jsonObject.findChildItemById("BT_screens", item.itemId, childItem.itemId))
                }
            }
        }

        BT_fileManager.saveTextFileToCache(configJson, BT_configuration.CACHED_CONFIGURATION_NAME)
        saveLastModified(config)

        return config
    }

    private fun saveLastModified(config: BtAppConfig) {
        val lastModifiedString = config.BT_items.firstOrNull()?.lastModified ?: return
        val lastModifiedObj = JSONObject().apply { put("lastModifiedUTC", lastModifiedString) }
        BT_fileManager.saveTextFileToCache(lastModifiedObj.toString(), BT_configuration.LAST_MODIFIED_NAME)
    }

    fun getCachedScreen(itemId: String): BT_item? = getCachedScreens().firstOrNull { it.itemId == itemId }

    fun getCachedScreens(): List<BT_item> = buzztouch_appDelegate.rootApp.screens

    suspend fun insert(item: BT_item) = withContext(Dispatchers.IO) { buzztouch_appDelegate.theContext?.let { BT_roomDatabase.insert(it, item) } }

    /**
     * Inserts BT_item objects into database
     */
    private suspend fun insertBtItemIntoDatabase(item: BT_item, index: Int, parentItemId: String?, type: BT_item.Type?, json: JSONObject?) {
        BtLog("insertBtItemIntoDatabase($type): ${item.toJson()}")
        val modifiedItem = item
            .copy(
                _id = null,
                image = null,
                hasChildItems = !item.childItems.isNullOrEmpty() || (item.jsonObject.getBooleanOrNull(KEY_CHILD_ITEMS) ?: false),
                itemIndex = index,
                parentItemGuid = parentItemId,
                type = type,
                jsonVars = json?.toString()
            )
        insert(modifiedItem)
    }
}
