package com.buzztouch.location

import android.Manifest
import android.app.Activity
import android.content.Context
import android.content.pm.PackageManager
import android.location.Location
import android.location.LocationListener
import android.location.LocationManager
import android.os.Bundle
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.extensions.BtLog
import com.buzztouch.extensions.canUseGps
import com.buzztouch.utils.*
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import kotlin.coroutines.CoroutineContext

object BT_locationManager : LocationListener, CoroutineScope {
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.IO

    // location listener...
    var locationManager: LocationManager? = null
    var locationUpdateCount = 0
    var locationListenerType = ""
    const val TAG = "BT_locationManager"
    private const val PERMISSION_PREF = "userAllowLocation"
    // private var mContext: Context? = null

    fun onPermissionResult(allowed: Boolean) {
        BT_preferencesKt.setBoolean(PERMISSION_PREF, allowed)
    }

    fun configureEnvironment(activity: Activity) {
        // location Manager. Remember the last reported location in the app's delegate so we don't have to turn on the GPS for every screen...
        if (buzztouch_appDelegate.foundUpdatedLocation) return
        if (!canUseGps) return
        if (!buzztouch_appDelegate.rootApp.startLocationUpdates) return BtLog("$TAG:configureEnvironment start GPS is set to NO (set to 0) in the applications configuration data, not starting GPS")

        BtLog("$TAG: start GPS is set to YES (set to 1) in the applications configuration data, trying to start GPS")
        val userAllowLocation = BT_preferencesKt.getBoolean(PERMISSION_PREF, true)
        if (!userAllowLocation) {
            BtLog("$TAG:configureEnvironment user has not prevented the GPS from starting using a BT_screen_settingsLocation screen")

            getLastGpsLocation(activity)
        } else {
            BtLog("$TAG:configureEnvironment user has prevented the GPS from starting using a BT_screen_settingsLocation screen")
        }
    } // remember in delegate

    private fun showPermissionsDialog(activity: Activity) {
        // Should we show an explanation?
        val requiresExplanation = ActivityCompat.shouldShowRequestPermissionRationale(activity, Manifest.permission.ACCESS_FINE_LOCATION)
        if (requiresExplanation) {
            // Show an explanation to the user.  After the user
            // sees the explanation, try again to request the permission.
            launch {
                withContext(Dispatchers.Main) { showRationale(activity) }
                requestPermission(activity)
            }
        } else {
            // No explanation needed, we can request the permission.
            requestPermission(activity)
        }
    }

    // TODO: Implement by developer if location is used by the app
    private suspend fun showRationale(context: Context) {
        // BT_alertUtils.showAlertAndWait(context, "Location Permission", "This app uses your location for map-related screens")
        BT_debugger.logError(TAG, "Requesting location permission - PLEASE IMPLEMENT BT_locationManager.showRationale()!!")
    }

    private fun requestPermission(activity: Activity) {
        ActivityCompat.requestPermissions(
            activity, arrayOf(Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION),
            BT_permissions.PERMISSION_CODE_LOCATION
        )
    }

    // s += " Updated: " + lastLocation.getTime();
    // start listening for location updates if we have GPS enabled...
    // only ask for location info "once" when app launches (saves battery)
    // getLastGPSLocation...

    fun getLastGpsLocation(context: Activity) {
        try {
            // only ask for location info "once" when app launches (saves battery)
            if (!buzztouch_appDelegate.foundUpdatedLocation) {
                locationUpdateCount = 0
                if (locationManager == null) {
                    locationUpdateCount = 0
                    locationManager = context.getSystemService(Context.LOCATION_SERVICE) as? LocationManager
                }
                locationManager?.let { locationManager ->
                    val permissionCheck = ContextCompat.checkSelfPermission(context, Manifest.permission.ACCESS_FINE_LOCATION)

                    // Here, thisActivity is the current activity
                    if (permissionCheck != PackageManager.PERMISSION_GRANTED) return showPermissionsDialog(context)

                    val lastLocation = locationManager.getLastKnownLocation(LocationManager.GPS_PROVIDER)
                    if (lastLocation != null) {

                        // remember in delegate
                        buzztouch_appDelegate.rootApp.rootDevice.deviceLatitude = lastLocation.latitude.toString()
                        buzztouch_appDelegate.rootApp.rootDevice.deviceLongitude = lastLocation.longitude.toString()
                        val logMsg = "Lat: ${lastLocation.latitude} Lon: ${lastLocation.longitude} Accuracy: ${lastLocation.accuracy}"
                        BtLog("$TAG:getLastGPSLocation $logMsg")
                    }
                    // start listening for location updates if we have GPS enabled...
                    locationListenerType = when {
                        locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER) -> LocationManager.GPS_PROVIDER
                        locationManager.isProviderEnabled(LocationManager.NETWORK_PROVIDER) -> LocationManager.NETWORK_PROVIDER
                        else -> ""
                    }
                    if (locationListenerType.isNotEmpty()) startListeningForLocationUpdate()
                    else BtLog("$TAG:getLastGPSLocation Can't start GPS or Network Services to get location info.")
                } ?: BtLog("$TAG:getLastGPSLocation locationManager == null?")
            }
        } catch (je: Exception) {
            BtLog("$TAG:getLastGPSLocation EXCEPTION $je")
        }
    }

    // LocationListener must implement these methods
    // onProviderDisabled (location manager provider)...
    override fun onProviderDisabled(theProvider: String) {
        BtLog("$TAG:onProviderDisabled The GPS is disabled on this device.")
    }

    // onProviderEnabled (location manager provider)...
    override fun onProviderEnabled(theProvider: String) {
        BtLog("$TAG:onProviderDisabled The GPS is enabled on this device.")
    }

    // onLocationChanged...
    override fun onLocationChanged(location: Location) {
        locationUpdateCount++
        BtLog("$TAG:onLocationChanged The device's location changed.")
        try {

            /*
  			 	Example of how to get the device's current location in any code you write anywhere in your app....

  				String myLatitude = buzztouch_appDelegate.rootApp.getRootDevice().getDeviceLatitude();
  				String myLongitude = buzztouch_appDelegate.rootApp.getRootDevice().getDeviceLongitude();
  				
  				Then convert myLatitude and myLongitude strings to doubles, integers, floats as needed.
  			 
  			*/
            buzztouch_appDelegate.rootApp.rootDevice.deviceLatitude = location.latitude.toString()
            buzztouch_appDelegate.rootApp.rootDevice.deviceLongitude = location.longitude.toString()

            val logMsg = "From: $locationListenerType Lat: ${location.latitude} Lon: ${location.longitude} Accuracy: ${location.accuracy}"
            BtLog("$TAG:onLocationChanged $logMsg")

            // stop listening after 10 reports (about 10 seconds) or if we have good accuracy faster....
            if (locationUpdateCount > 10 || location.accuracy < 25) {
                BtLog("$TAG:onLocationChanged turning off GPS to save battery, saved last location.")

                // flag foundUpdatedLocation in the delgate so other screens don't turn on the GPS
                buzztouch_appDelegate.foundUpdatedLocation = true

                // stop listening (kill the locationManager)...
                stopListeningForLocationUpdate()
            }
        } catch (e: Exception) {
        }
    }

    // onStatusChanged...
    override fun onStatusChanged(theProvider: String, status: Int, extras: Bundle) {
        try {
            BtLog("$TAG:onStatusChanged (for the location manager)")
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    // startListeningForLocationUpdate..
    @Throws(SecurityException::class)
    fun startListeningForLocationUpdate() {
        BtLog("$TAG:startListeningForLocationUpdate (started listening for location changes)")
        try {
            locationManager?.let { locationManager ->
                // we we started this in getLastGPSLocation() we set a flag to tell us what type of service to setup..
                locationManager.requestLocationUpdates(locationListenerType, 0, 0f, this)
                BtLog("$TAG:startListeningForLocationUpdate asking for locations updates from $locationListenerType...")
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    @Throws(SecurityException::class)
    fun stopListeningForLocationUpdate() {
        BtLog("$TAG:stopListeningForLocationUpdate (stopped listening to location changes)")
        try {
            locationManager?.removeUpdates(this)
            locationManager = null
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }
}
