/*  File Version: 4.0.1
 *	Copyright, Chris Ruddell, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the
 *	following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list
 *	of conditions and the following disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors
 *	may be used to endorse or promote products derived from this software without specific
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 *	OF SUCH DAMAGE.
 */
package com.buzztouch.push_notifications

import android.app.Notification
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.Intent
import androidx.core.app.NotificationCompat
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.BuildConfig
import com.buzztouch.R
import com.buzztouch.ui.bt_screens.BT_activity_host
import com.buzztouch.utils.BT_debugger
import com.buzztouch.utils.BT_preferences
import com.google.firebase.messaging.FirebaseMessagingService
import com.google.firebase.messaging.RemoteMessage
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlin.coroutines.CoroutineContext

// Created by Christopher Ruddell, 12/28/2016
class BT_FirebaseMessagingService : FirebaseMessagingService(), CoroutineScope {
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.IO

    override fun onNewToken(token: String) {
        super.onNewToken(token)

        // ask BT_gcmServerUtils to send googleCloudMessagingToken to backend server...
        launch {
            BT_preferences.setPrefString("googleCloudMessagingToken", token)
            BT_gcmServerUtilsKt.gcmRegisterOnServer(token)
        }
    }

    override fun onMessageReceived(remoteMessage: RemoteMessage) {
        super.onMessageReceived(remoteMessage)
        val notification: RemoteMessage.Notification? = remoteMessage.notification
        val dataMap: Map<String, String> = remoteMessage.data ?: mapOf()
        if (notification != null) {
            val messageTitle = remoteMessage.notification!!.title
            val messageBody = remoteMessage.notification!!.body
            val message = messageTitle ?: (messageBody ?: "")
            BT_debugger.logInfo(this, "onMessageReceived:$message")
            sendNotification(message)
            //            showNotification(message);
        } else if (dataMap.isNotEmpty()) {
            BT_debugger.logInfo(this, "data:")
            val message: String = dataMap["message"] ?: ""
            val sound: String = dataMap["sound"] ?: ""
            BT_debugger.logInfo(this, "onMessageReceived:$message")
            sendNotification(message)
            //            showNotification(message);
        }
    }

    // Put the message into a notification and post it.
    private fun sendNotification(msg: String?) {
        BT_debugger.logInfo(this, "BT_FirebaseMessagingService:sendNotification: $msg")
        val mNotificationManager = this.getSystemService(NOTIFICATION_SERVICE) as NotificationManager
        val contentIntent = PendingIntent.getActivity(this, 0, Intent(this, BT_activity_host::class.java), 0)
        val mBuilder = NotificationCompat.Builder(this)
            .setSmallIcon(R.mipmap.ic_launcher)
            .setContentTitle(this.getString(R.string.app_name))
            .setStyle(NotificationCompat.BigTextStyle().bigText(msg))
            .setPriority(Notification.PRIORITY_MAX)
            .setFullScreenIntent(contentIntent, true)
            .setContentText(msg)
        mBuilder.setContentIntent(contentIntent)
        mNotificationManager.notify(NOTIFICATION_ID, mBuilder.build())
    }

    /**
     * Shows the message as an alert dialog popup if the app is currently in the foreground.
     * If using this method, it is recommended that you comment out "setFullScreenIntent()" in the sendNotificaiton() method above.
     * @param message
     */
    private fun showNotification(message: String) {
        val intent = Intent()
        intent.action = ACTION_PUSH_NOTIFICATION
        intent.putExtra(ARG_NOTIFICATION_MESSAGE, message)
        sendBroadcast(intent)
    }

    companion object {
        const val NOTIFICATION_ID = 1001
        const val ARG_NOTIFICATION_MESSAGE = "bt_notification_message"
        const val ACTION_PUSH_NOTIFICATION = BuildConfig.APPLICATION_ID + ".BT_PUSH_NOTIFICATION"
    }
}
