/*  File Version: 4.0.1
 *	Copyright, Chris Ruddell, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.push_notifications

import android.net.Uri
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.extensions.*
import com.buzztouch.utils.BT_preferences
import com.buzztouch.utils.BT_preferencesKt
import com.buzztouch.utils.BT_strings
import org.json.JSONObject

object BT_gcmServerUtilsKt {
    var serverURL = buzztouch_appDelegate.rootApp.registerForPushURL
    var postCommand: String? = null
    var tmpDeviceToken: String? = null

    // setupPostValues...
    fun setupPostValues(): HashMap<String, String> {

        // build URL...
        if (serverURL.length > 1) {
            serverURL = BT_strings.mergeBTVariablesInString(serverURL)
            serverURL = BT_strings.encodeCompleteURL(serverURL)
            try {

                // pull apiKey and apiSecret from URL params...
                val uri = Uri.parse(serverURL)
                var useApiKey = uri.getQueryParameter("apiKey")
                var useApiSecret = uri.getQueryParameter("apiSecret")
                if (useApiKey == null) useApiKey = ""
                if (useApiSecret == null) useApiSecret = ""

                // create encoded form values
                val encodedFormValues = hashMapOf(
                    "command" to "registerForPush",
                    "gcmCommand" to "",
                    "appGuid" to useApiKey,
                    "apiSecret" to useApiSecret,
                    "deviceId" to buzztouch_appDelegate.rootApp.rootDevice.deviceId,
                    "deviceLatitude" to deviceLatitude,
                    "deviceLongitude" to deviceLongitude,
                    "deviceModel" to BT_strings.urlEncode(buzztouch_appDelegate.rootApp.rootDevice.deviceModel),
                    "userId" to buzztouch_appDelegate.rootApp.rootUser.userId,
                    "deviceType" to "android",
                    "deviceToken" to (tmpDeviceToken ?: ""),
                    "currentMode" to buzztouch_appDelegate.rootApp.currentMode
                )

                return encodedFormValues
            } catch (e: Exception) {
                e.printStackTrace()
            }
        }

        return hashMapOf()
    }

    // gcmRegisterOnServer...
    suspend fun gcmRegisterOnServer(regId: String?) {
        BtLog("BT_gcmServerUtils:gcmRegisterOnServer")

        // set tmpDevice token to the value passed in...
        tmpDeviceToken = regId

        // setup post values...
        val formValues = setupPostValues()
        if (serverURL.length > 1) {
            formValues["gcmCommand"] = "registerDevice"
            val response: String? = formValues.postToUrl<String>(serverURL)
            BtLog("gcmRegisterOnServer: $response")
            processResults(response)
        }
    }

    // gcmUnregisterOnServer...
    suspend fun gcmUnregisterOnServer(regId: String?) {
        BtLog("BT_gcmServerUtils:gcmUnregisterOnServer")

        // set tmpDevice token to the value passed in...
        tmpDeviceToken = regId

        // setup post values...
        val formValues = setupPostValues()

        // must have URL...
        if (serverURL.length > 1) {
            formValues["gcmCommand"] = "unregisterDevice"
            val response: String? = formValues.postToUrl<String>(serverURL)
            BtLog("gcmUnregisterOnServer: $response")
            processResults(response)
        }
    }

    private fun processResults(results: String?) {
        // parse JSON data...
        if (results.isNullOrEmpty()) return BtLog("BT_gcmServerUtils:ServerAsyncTask The server did not return a result?")
        try {
            // raw JSON results (entire string returned by server)...
            val objRaw = JSONObject(results)
            val result: JSONObject = objRaw.getJsonObjectOrNull("result") ?: return
            val status = result.getStringOrNull("result") ?: return BtLog("BT_gcmServerUtils: no result found")
            if (!status.equalsIgnoreCase("valid")) return BtLog("BT_gcmServerUtils: JSON results returned \"failed\" status")
            val deviceToken = result.getStringOrNull("deviceToken") ?: return BtLog("BT_gcmServerUtils: deviceToken not found")
            BtLog("BT_gcmServerUtils:ServerAsyncTask Device Token: $deviceToken")

            when (postCommand) {
                "registerDevice" -> onRegisterDevice(deviceToken)
                "unregisterDevice" -> onUnregisterDevice()
            }
        } catch (e: Exception) {
            BtLog("BT_gcmServerUtils:ServerAsyncTask EXCEPTION parsing JSON returned from server: $e")
        }
    }

    private fun onRegisterDevice(deviceToken: String) {
        BtLog("BT_gcmServerUtils:ServerAsyncTask DEVICE REGISTERED")

        // save deviceToken (GCM Registration Id) in prefs and delegate...
        BT_preferencesKt.setString("googleCloudMessagingToken", deviceToken)
        buzztouch_appDelegate.rootApp.pushRegistrationId = deviceToken
    }

    private fun onUnregisterDevice() {
        BtLog("BT_gcmServerUtils:ServerAsyncTask DEVICE UNREGISTERED")

        // remove deviceToken (GCM Registration Id) from prefs and delegate...
        BT_preferencesKt.setString("googleCloudMessagingToken", "")
        buzztouch_appDelegate.rootApp.pushRegistrationId = ""
    }
}
