/*  File Version: 4.0.1
 *	Copyright, Chris Ruddell, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the
 *	following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list
 *	of conditions and the following disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors
 *	may be used to endorse or promote products derived from this software without specific
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 *	OF SUCH DAMAGE.
 */
package com.buzztouch.push_notifications

import android.app.Activity
import android.content.Context
import android.content.DialogInterface
import android.widget.Toast
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.R
import com.buzztouch.extensions.BtLog
import com.buzztouch.utils.*
import com.google.android.gms.common.ConnectionResult
import com.google.android.gms.common.GoogleApiAvailability
import com.google.firebase.messaging.FirebaseMessaging
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import kotlin.coroutines.CoroutineContext

object BT_pushManager : CoroutineScope {
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.IO
    var googleCloudMessagingToken: String = ""
    var promptForPush = false

    private const val TAG = "BT_pushManager"
    private const val DEBUG_LOG = true

    // ///////////////////////////////////////////////////
    // Google Cloud Messaging methods (Push Notifications)
    fun configureEnvironment(activity: Activity) {
        // register for push logic...
        if (buzztouch_appDelegate.rootApp.promptForPushNotifications == "1" && buzztouch_appDelegate.googleCloudMessagingProjectNumber.length > 1) {
            if (buzztouch_appDelegate.rootApp.registerForPushURL.length > 3) {
                BT_debugger.showIt(TAG + ":configureEnvironment Prompt For Push Notifications is ON (set to 1).")
                promptForPush = true
            }
        } else {
            BT_debugger.showIt(TAG + ":configureEnvironment Prompt For Push Notifications is OFF (set to 0) in app's configuration data or AndroidManifest does not have a GCM Project Number")
        }

        // if setup for push...
        if (promptForPush) {

            // reference Google Cloud Messaging if available...
            if (checkPlayServices(activity)) {

                // all done if the googleCloudMessagingToken is saved in the preferences...
                googleCloudMessagingToken = BT_preferencesKt.getString("googleCloudMessagingToken") ?: ""

                // prompt user for "allow" push notifications if they have NOT previously said "no thanks"...
                if (googleCloudMessagingToken.isEmpty()) {
                    if (BT_fileManager.doesCachedFileExist("rejectedpush.txt")) {
                        BT_debugger.showIt(TAG + ":configureEnvironment Device owner has rejected Push Notifications")
                    } else {
                        launch { confirmRegisterForPush(activity.applicationContext) }
                    }
                } else {
                    BT_debugger.showIt(TAG + ":configureEnvironment Device is already registered for GCM (Push) with GCM Id: " + googleCloudMessagingToken)
                }
            } else {
                Toast.makeText(activity, "Your device cannot recieve push notifications. Google Play Services not enabled on device", Toast.LENGTH_SHORT).show()
            }
        }
    }

    // confirmRegisterForPush...
    suspend fun confirmRegisterForPush(context: Context) {
        val confirmPush: Int = withContext(Dispatchers.Main) {
            BT_alertUtils.showConfirm(
                context,
                "Accept Push Notifications?",
                "This app would like to send you notifications and simple messages. Is this OK?",
                R.mipmap.ic_launcher
            )
        }
        when (confirmPush) {
            DialogInterface.BUTTON_POSITIVE -> registerForPush()
            DialogInterface.BUTTON_NEGATIVE -> BT_fileManager.saveTextFileToCache("rejected", "rejectedpush.txt")
        }
    }

    // registerForPush...
    suspend fun registerForPush() {
        try {
            // do we have GCM Project number?...
            if (buzztouch_appDelegate.googleCloudMessagingProjectNumber.isNotEmpty()) {
                FirebaseMessaging.getInstance().token.addOnCompleteListener { task ->

                }
                googleCloudMessagingToken = getDeviceToken() ?: ""
                BtLog("$TAG:RegisterForPushAsync Registration ID: $googleCloudMessagingToken")

                // ask BT_gcmServerUtils to send googleCloudMessagingToken to backend server...
                if (googleCloudMessagingToken.isNotEmpty()) BT_gcmServerUtilsKt.gcmRegisterOnServer(googleCloudMessagingToken)
            }
        } catch (e: Exception) {
            e.printStackTrace()
            BtLog("$TAG : registerForPush ${e.localizedMessage}")
        }
    }

    suspend fun getDeviceToken(): String? = safeSuspendCoroutine { cont ->
        FirebaseMessaging.getInstance().token.addOnCompleteListener { task ->
            if (!task.isSuccessful) {
                BtLog("Fetching FCM registration token failed")
                cont.resume(null)
                return@addOnCompleteListener
            }

            cont.resume(task.result)
        }
    }

    // ///////////////////////////////////////////////////
    // unregisterForPush...
    suspend fun unregisterForPush() {
        try {

            // do we have GCM Project number?...
            if (buzztouch_appDelegate.googleCloudMessagingProjectNumber.length > 1) {
                BT_debugger.showIt(TAG + ":UnRegisterForPushAsync Registration ID: " + googleCloudMessagingToken)

                // ask BT_gcmServerUtils to unregister googleCloudMessagingToken on backend server...
                BT_gcmServerUtilsKt.gcmUnregisterOnServer(googleCloudMessagingToken)
            }
        } catch (e: Exception) {
            BT_debugger.showIt(TAG + ":registerForPush: " + e.toString())
        }
    }

    // checkPlayServices...
    fun checkPlayServices(activity: Activity): Boolean {
        val resultCode = GoogleApiAvailability.getInstance().isGooglePlayServicesAvailable(activity)
        if (resultCode != ConnectionResult.SUCCESS) {
            if (GoogleApiAvailability.getInstance().isUserResolvableError(resultCode)) {
                GoogleApiAvailability.getInstance().getErrorDialog(activity, resultCode, 9000)?.show()
            } else {
                activity.finish()
            }
            return false
        }
        return true
    }
}
