/*
 *	Copyright 2011, David Book, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.ui.bt_screens

import com.buzztouch.R
import com.buzztouch.extensions.BtLog
import com.buzztouch.extensions.canSendEmail
import com.buzztouch.extensions.copyAssetToCache
import com.buzztouch.extensions.copyCachedFileToSdCard
import com.buzztouch.extensions.doesCachedFileExist
import com.buzztouch.extensions.projectAssetExists
import com.buzztouch.extensions.propertyValue
import com.buzztouch.extensions.readCachedFile
import com.buzztouch.extensions.readTextFromUrl
import com.buzztouch.extensions.shareFileViaEmail
import com.buzztouch.extensions.shareUrlViaEmail
import com.buzztouch.extensions.showAlert
import com.buzztouch.extensions.showProgress
import com.buzztouch.utils.BT_fileManager
import kotlinx.coroutines.launch

class BT_screen_customHTML : BT_screen_customURL() {
    var localFileName = ""
    var saveAsFileName = ""
    var forceRefresh = false

    override fun setInitialWebProperties() {
        super.setInitialWebProperties()
        dataURL = screenData.propertyValue("dataURL")
        localFileName = screenData.propertyValue("localFileName")
        forceRefresh = (screenData.propertyValue("forceRefresh", "0").toIntOrNull() ?: 0) == 1
        currentURL = dataURL
        originalURL = dataURL
    }

    override fun loadInitialPage() {
        // if we have NO localFileName and no dataURL, use the sample file for this plugin...
        if (localFileName.isEmpty() && dataURL.isEmpty()) {
            localFileName = "bt_screen_customhtml_sample.html"
        }

        // setup the saveAsFileName
        saveAsFileName = when {
            localFileName.isNotEmpty() -> localFileName.also { saveAsFileName.copyAssetToCache() }
            else -> screenData.itemId + "_screenData.html"
        }

        BtLog("loadInitialPage(forceFresh:$forceRefresh): $saveAsFileName; $dataURL")

        // remove file if we are force-refreshing...
        if (forceRefresh) BT_fileManager.deleteFile(saveAsFileName)

        when {
            dataURL.isEmpty() -> loadFromLocalFile(saveAsFileName)
            else -> loadFromUrlOrCache(dataURL, saveAsFileName)
        }
    }

    private fun loadFromUrlOrCache(url: String, cacheName: String) {
        // if we have a cached version, load that...
        if (forceRefresh || !cacheName.doesCachedFileExist()) {
            // load from URL
            BtLog("$fragmentName: loading from URL: $url")
            downloadAndSaveFile(url, saveAsFileName)
        } else {
            // load from cache
            BtLog("$fragmentName: loading from cache: $cacheName")
            val theData = cacheName.readCachedFile()
            showProgress()
            loadDataString(theData)
        }
    }

    private fun loadFromLocalFile(fileName: String) {
        BtLog("$fragmentName: loadFromLocalFile($fileName)")
        // HTML doc must be in /assets/BT_Docs folder...
        if (fileName.projectAssetExists("BT_Docs")) {
            // load from BT_Docs folder...
            BtLog("$fragmentName: loading from BT_Docs: $fileName")
            showProgress()
            webView?.loadUrl("file:///android_asset/BT_Docs/$fileName")
        } else {
            BtLog("$fragmentName: ERROR. HTML file \"$fileName\" does not exist in BT_Docs folder and not URL found? Not loading.")
            showAlert(getString(R.string.errorTitle), getString(R.string.errorLoadingScreen))
        }
    }

    // load html string...
    private fun loadDataString(theString: String?) {
        BtLog("$fragmentName: loadDataString")
        webView?.loadDataWithBaseURL(null, theString ?: getString(R.string.errorLoadingScreen), "text/html", "utf-8", "about:blank")
        hideProgress()
    }

    // handleEmailDocumentButton
    override fun handleEmailDocumentButton() {
        if (!canSendEmail) {
            showAlert(R.string.noNativeAppTitle, R.string.noNativeAppDescription)
            BtLog("$fragmentName:handleEmailDocumentButton Cannot email document, no URL provided")
            return
        }

        if (!saveAsFileName.doesCachedFileExist()) {
            showAlert(R.string.cannotEmailDocument, R.string.cannotOpenDocument)
            BtLog("$fragmentName:handleEmailDocumentButton Cannot email document, document not saved")
            return
        }

        try {
            val file = saveAsFileName.copyCachedFileToSdCard() ?: return BtLog("$fragmentName:handleEmailDocumentButton Cannot email document, file does not exist: $saveAsFileName")

            if (currentURL.isNotEmpty()) shareUrlViaEmail(currentURL)
            else shareFileViaEmail(file)
        } catch (e: Exception) {
            showAlert(R.string.errorTitle, R.string.cannotEmailDocument)
            BtLog("$fragmentName:handleEmailDocumentButton EXCEPTION $e")
        }
    }

    private fun downloadAndSaveFile(dataURL: String?, saveAsFileName: String) = launch {
        showProgress()
        when (saveAsFileName.doesCachedFileExist()) {
            true -> loadDataString(saveAsFileName.readCachedFile())
            false -> dataURL?.readTextFromUrl(saveAsFileName)
        }
        hideProgress()
    }
}
