/*
 *	Copyright 2011, David Book, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.ui.bt_screens

import android.content.DialogInterface
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.webkit.WebResourceRequest
import android.webkit.WebView
import android.webkit.WebViewClient
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.R
import com.buzztouch.extensions.*
import com.buzztouch.utils.BT_alertUtils
import com.buzztouch.utils.BT_fileManager
import kotlinx.coroutines.launch

open class BT_screen_customURL : BT_fragment() {
    var webView: WebView? = null
    var dataURL = ""
    var currentURL = ""
    var originalURL = ""
    val showBrowserBarBack: Boolean get() = screenData.booleanValue("showBrowserBarBack") ?: true
    val showBrowserBarLaunchInNativeApp: Boolean get() = screenData.booleanValue("showBrowserBarLaunchInNativeApp") ?: true
    val showBrowserBarEmailDocument: Boolean get() = screenData.booleanValue("showBrowserBarEmailDocument") ?: true
    val showBrowserBarRefresh: Boolean get() = screenData.booleanValue("showBrowserBarRefresh") ?: true
    private var didInit = false

    companion object {
        private const val MENU_BACK = 1
        private const val MENU_LAUNCH_IN_NATIVE_APP = 2
        private const val MENU_EMAIL_DOCUMENT = 3
        private const val MENU_REFRESH = 4
    }

    // onCreateView...
    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View? {

        /*
 			Note: fragmentName property is already setup in the parent class (BT_fragment). This allows us 
 			to add the 	name of this class file to the LogCat console using the BT_debugger.
		*/
        // show life-cycle event in LogCat console...
        if (screenData == null) return null
        BtLog(fragmentName + ":onCreateView JSON itemId: \"" + screenData.itemId + "\" itemType: \"" + screenData.itemType + "\" itemNickname: \"" + screenData.itemNickname + "\"")

        // inflate the layout file for this screen...
        val thisScreensView = inflater.inflate(R.layout.bt_screen_customurl, container, false)

        // reference to the webview in the layout file.
        webView = thisScreensView.findViewById(R.id.webView)
        webView?.apply {
            setBackgroundColor(0)
            setInitialScale(0)
            settings.apply {
                javaScriptEnabled = true
                setSupportZoom(true)
                builtInZoomControls = true
            }
        }

        webView?.webViewClient = webClient()

        setInitialWebProperties()

        // prevent user interaction? - preventUserInteraction property not supported on Android currently

        // hide scroll bars..
        if (booleanValue("hideVerticalScrollBar")) webView?.disableVerticalScrolling()
        if (booleanValue("hideHorizontalScrollBar")) webView?.disableHorizontalScrolling()
        if (booleanValue("preventAllScrolling")) {
            webView?.disableVerticalScrolling()
            webView?.disableHorizontalScrolling()
        }

        loadInitialPage()

        // return...
        return thisScreensView
    } // onCreateView...

    private fun WebView.disableVerticalScrolling() { this.isVerticalScrollBarEnabled = false }
    private fun WebView.disableHorizontalScrolling() { this.isHorizontalScrollBarEnabled = false }

    open fun webClient() = BtWebClient()

    open fun setInitialWebProperties() {
        // fill JSON properties...
        dataURL = screenData.propertyValue("dataURL")

        // must have a dataURL...
        if (dataURL.isEmpty()) {
            dataURL = "https://www.google.com"
        }
        currentURL = dataURL
        originalURL = dataURL
        if (!dataURL.contains("http")) {
            // most likely a local file.  Get URI to file from Assets directory
            if (BT_fileManager.doesProjectAssetExist("BT_Docs", dataURL)) {
                // yep - its a local file - we found it!
                // note - this path should only be used by WebView and not taken as a literal path to the file
                dataURL = "file:///android_asset/BT_Docs/$dataURL"
            }
        }
    }

    // load URL in webView
    private fun loadUrl(theUrl: String?) = runSafe("$fragmentName: loadUrl") { theUrl?.let { webView?.loadUrl(theUrl) } }

    open fun loadInitialPage() {
        // figure out what to load...
        if (dataURL.isNotEmpty()) {
            BtLog("$fragmentName: loading URL from: ${dataURL.mergeBtVariables()}")
            loadUrl(dataURL.mergeBtVariables())
        } else {
            BtLog("$fragmentName: No URL found? Not loading web-view!")
            showAlert(getString(R.string.errorTitle), getString(R.string.errorNoURL))
        }
    }

    // back button...
    override fun handleBackButton() {
        BtLog("$fragmentName:handleBackButton")
        when (webView?.canGoBack()) {
            true -> webView?.goBack()
            else -> {
                showToast(R.string.errorNoHistory)
                BtLog("$fragmentName:handleBackButton cannot go back?")
            }
        }
    }

    // refresh button...
    private fun handleRefreshButton() {
        BtLog("$fragmentName:handleRefreshButton Current URL: $currentURL")
        if (currentURL.isNotEmpty()) {
            showProgress()
            webView?.loadUrl(currentURL)
        } else {
            BtLog("$fragmentName:handleRefreshButton cannot refresh URL: $currentURL")
        }
    }

    // launch in native app button...
    private fun handleLaunchInNativeAppButton() {
        BtLog("$fragmentName:handleLaunchInNativeAppButton")
        if (currentURL.isNotEmpty() && originalURL.isNotEmpty()) {
            confirmLaunchInNativeApp()
        } else {
            showAlert(R.string.errorTitle, R.string.cannotOpenDocumentInNativeApp)
            BtLog("$fragmentName:handleLaunchInNativeAppButton NO url?")
        }
    }

    // handle email button
    open fun handleEmailDocumentButton() {
        BtLog("$fragmentName:handleLaunchInNativeAppButton")

        // must be able to send email...
        when {
            !canSendEmail -> {
                showAlert(getString(R.string.noNativeAppTitle), getString(R.string.noNativeAppDescription))
                BtLog("$fragmentName:handleEmailDocumentButton Cannot email document, device doesn't support email")
            }
            currentURL.isEmpty() -> {
                showAlert(getString(R.string.errorTitle), getString(R.string.cannotEmailDocument))
                BtLog("$fragmentName:handleEmailDocumentButton Cannot email document, URL not available")
            }
            else -> launchEmailApp(currentURL)
        }
    }

    private fun launchEmailApp(url: String) = try {
        // tell Android launch the native email application...
        val emailIntent = Intent(Intent.ACTION_SEND)
        emailIntent.type = "plain/text"
        emailIntent.putExtra(Intent.EXTRA_SUBJECT, getString(R.string.sharingWithYou))
        emailIntent.putExtra(Intent.EXTRA_TEXT, url)

        // chooser will prompt user if they have more than one email client..
        startActivity(Intent.createChooser(emailIntent, getString(R.string.openWithWhatApp)))
    } catch (e: Exception) {
        showAlert(getString(R.string.errorTitle), getString(R.string.cannotEmailDocument))
        BtLog("$fragmentName:handleEmailDocumentButton EXCEPTION $e")
    }

    // confirm launch in native app
    private fun confirmLaunchInNativeApp() {
        val context = context ?: return
        launch {
            when (context.showConfirm(R.string.confirm, R.string.confirmLaunchInNativeBrowser)) {
                DialogInterface.BUTTON_POSITIVE -> {
                    try {
                        val i = Intent(Intent.ACTION_VIEW, Uri.parse(currentURL))
                        startActivity(i)
                    } catch (e: Exception) {
                        BtLog("$fragmentName: Error launching native app for url: $currentURL")
                        showAlert(buzztouch_appDelegate.getApplication().getString(R.string.noNativeAppTitle), getString(R.string.noNativeAppDescription))
                    }
                }
            }
        }
    }

    // confirm email document
    fun confirmEmailDocumentDialogue() {
        val context = context ?: return
        launch {
            val selection = BT_alertUtils.showConfirm(context, getString(R.string.confirm), getString(R.string.confirmEmailDocument), R.drawable.bt_blank)
            when (selection) {
                DialogInterface.BUTTON_POSITIVE -> {
                    // tell Android launch the native email application...
                    val emailIntent = Intent(Intent.ACTION_SEND)
                    emailIntent.type = "plain/text"
                    emailIntent.putExtra(Intent.EXTRA_SUBJECT, getString(R.string.sharingWithYou))
                    emailIntent.putExtra(Intent.EXTRA_TEXT, currentURL)

                    // chooser will propmpt user if they have more than one email client..
                    startActivity(Intent.createChooser(emailIntent, getString(R.string.openWithWhatApp)))
                }
            }
        }
    }

    // onCreateOptionsMenu...
    override fun onCreateOptionsMenu(menu: Menu, inflater: MenuInflater) {
        BtLog("$fragmentName:onCreateOptionsMenu JSON itemId: \"$screenItemId\"")

        /*
	 	 	Screens that extend BT_fragment should implement their own onCreateOptionsMenu() method to add
	 	 	action bar items specific to that screen. They should call super.onCreateOptionsMenu() AFTER
	 	 	adding customized items for that screen. 	  
	 	*/

        // for individual menu items...
        var menuItem: MenuItem? = null

        // back...
        if (showBrowserBarBack) {
            menuItem = menu.add(Menu.NONE, MENU_BACK, 0, getString(R.string.back))
            menuItem.icon = BT_fileManager.getDrawableByName("bt_arrow_left.png")
            menuItem.setShowAsAction(MenuItem.SHOW_AS_ACTION_IF_ROOM)
        }

        // launch in native app...
        if (showBrowserBarLaunchInNativeApp) {
            menuItem = menu.add(Menu.NONE, MENU_LAUNCH_IN_NATIVE_APP, 0, getString(R.string.browserOpenInNativeApp))
            menuItem.icon = BT_fileManager.getDrawableByName("bt_web_site.png")
            menuItem.setShowAsAction(MenuItem.SHOW_AS_ACTION_IF_ROOM)
        }

        // email document...
        if (showBrowserBarEmailDocument) {
            menuItem = menu.add(Menu.NONE, MENU_EMAIL_DOCUMENT, 0, getString(R.string.browserEmailDocument))
            menuItem.icon = BT_fileManager.getDrawableByName("bt_email.png")
            menuItem.setShowAsAction(MenuItem.SHOW_AS_ACTION_IF_ROOM)
        }

        // refresh page...
        if (showBrowserBarRefresh) {
            menuItem = menu.add(Menu.NONE, MENU_REFRESH, 0, getString(R.string.browserRefresh))
            menuItem.icon = BT_fileManager.getDrawableByName("bt_refresh.png")
            menuItem.setShowAsAction(MenuItem.SHOW_AS_ACTION_IF_ROOM)
        }

        // call super...
        super.onCreateOptionsMenu(menu, inflater)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        BtLog(fragmentName + ":onOptionsItemSelected JSON itemId: \"" + screenItemId + "\" Selected Item's Id: " + item.itemId)
        return when (item.itemId) {
            MENU_BACK -> { handleBackButton(); true }
            MENU_LAUNCH_IN_NATIVE_APP -> { handleLaunchInNativeAppButton(); true }
            MENU_EMAIL_DOCUMENT -> { handleEmailDocumentButton(); true }
            MENU_REFRESH -> { handleRefreshButton(); true }
            else -> super.onOptionsItemSelected(item)
        }
    }

    inner class BtWebClient : WebViewClient() {
        override fun shouldOverrideUrlLoading(view: WebView, request: WebResourceRequest): Boolean {
            val url = request.url.toString()

            return when (canLoadDocumentInWebView(url)) {
                true -> false.also { currentURL = url; showProgress() }
                false -> true.also {
                    currentURL = url
                    openUrlWithApp(url)
                }
            }
        }

        private fun openUrlWithApp(url: String) {
            // ask user what app to open this in if the method returned NO...
            try {
                val i = Intent(Intent.ACTION_VIEW, Uri.parse(url))
                startActivity(Intent.createChooser(i, getString(R.string.openWithWhatApp)))
            } catch (e: Exception) {
                BtLog("$fragmentName: Error launching native app for url: $url")
                showAlert(R.string.noNativeAppTitle, R.string.noNativeAppDescription)
            }
        }

        override fun onPageFinished(view: WebView, url: String) {
            hideProgress()
            BtLog("$fragmentName:onPageFinished finished Loading: $url")
        }

        override fun onReceivedError(view: WebView, errorCode: Int, description: String, failingUrl: String) {
            hideProgress()
            showAlert(R.string.errorTitle, R.string.errorLoadingScreen)
            BtLog(":onReceivedError ERROR loading url: " + failingUrl + " Description: " + description.also { fragmentName = it })
        }
    }
}
