/*  File Version: 4.0.4
 *	Copyright, David Book, buzztouch.com
 *  Updates made by Chris Ruddell, 2/4/2017
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.ui.bt_screens

import android.content.Intent
import android.os.Bundle
import android.os.Handler
import android.os.Message
import android.view.LayoutInflater
import android.view.MotionEvent
import android.view.View
import android.view.View.OnTouchListener
import android.view.ViewGroup
import android.view.animation.AnimationUtils
import android.widget.FrameLayout
import android.widget.ImageView
import android.widget.LinearLayout
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.R
import com.buzztouch.data_models.BT_item
import com.buzztouch.extensions.*
import com.buzztouch.utils.BT_debugger
import com.buzztouch.utils.BT_strings
import com.buzztouch.utils.BT_viewUtilities
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch

class BT_screen_splash : BT_fragment(), OnTouchListener {
    // properties...
    private var transitionType = ""
    private var effectName = ""
    private var startTransitionAfterSeconds = 0
    private var transitionDurationSeconds = 0

    // onCreateView...
    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View? {
        BtLog("$fragmentName:onCreateView JSON itemId: ${screenData.itemId} itemType: ${screenData.itemType} itemNickname: ${screenData.itemNickname}")

        // inflate the layout file for this screen...
        val thisScreensView = LinearLayout(inflater.context).apply {
            setSize(MATCH_PARENT, MATCH_PARENT)
            setOnTouchListener(this@BT_screen_splash)
        }

        // must have screen data...
        screenData ?: let {
            onScreenDataError()
            return thisScreensView
        }
        // ask BT_viewUtilities to setup this screens background color...
        BT_viewUtilities.updateBackgroundColorsForScreen(this.activity, screenData)

        // get animation values from JSON...
        transitionType = propertyValue("transitionType")
        effectName = propertyValue("effectName", "fallDown")
        startTransitionAfterSeconds = intValue("startTransitionAfterSeconds", 1)
        transitionDurationSeconds = intValue("transitionDurationSeconds", 1)

        // if we have an effect we do it first, then start the animation sequence...
        if (effectName.isNotEmpty()) {
            view?.postDelayed(
                { // pass effect type and reference to the background image to the startEffect method..
                    val imgView = activity?.findViewById<View>(R.id.backgroundImageView) as ImageView
                    startEffect(effectName, imgView)
                },
                100
            )
        }

        // return...
        return thisScreensView
    }

    private fun onScreenDataError() {
        // show a message if this fragment is NOT running in the BT_activity_start Activity... (it should be)...
        if (activity !is BT_activity_start) showToast("Note: The Action Bar will be hidden when splash screens are used during app launch.", "long")
    }

    // onResume...
    override fun onResume() {
        super.onResume()

        // will hold the name and/or URL of the randomized image...
        var tmpImage = propertyValue(if (isLargeDevice) "backgroundImageNameLargeDevice" else "backgroundImageNameSmallDevice")
        var tmpURL = propertyValue(if (isLargeDevice) "backgroundImageURLLargeDevice" else "backgroundImageURLSmallDevice")

        // use a default image (included in the BT core) if we didn't find one in the childItems...
        if (tmpImage.isEmpty() && tmpURL.isEmpty()) {
            tmpImage = "bt_member_275.png"
            tmpURL = ""
        }

        // create a BT_item to pass to the updateBackgroundImageForScreen() method in BT_viewUtilities...
        val tmpItem = BT_item()
        tmpItem.itemId = screenData.itemId
        tmpItem.itemType = screenData.itemType
        tmpItem.itemNickname = screenData.itemNickname

        // use the backgroundImageScale property for the screen...
        val backgroundImageScale = BT_strings.getJsonPropertyValue(screenData.jsonObject, "backgroundImageScale", "center")

        // pass the same image for both small and large properties (we figured it out already)...
        val tmpJson = """
            {
                "backgroundImageNameLargeDevice":"$tmpImage",
                "backgroundImageNameSmallDevice":"$tmpImage",
                "backgroundImageURLLargeDevice":"$tmpURL",
                "backgroundImageURLSmallDevice":"$tmpURL",
                "backgroundImageScale":"$backgroundImageScale"
            }
        """.trimIndent()

        // create JSONObject from string using BT_strings class...				
        tmpItem.jsonObject = BT_strings.getJsonObjectFromString(tmpJson)

        // ask BT_viewUtilities to update our background image view using this randomized JSON data...
        val backgroundImageView = this.activity?.findViewById<View>(R.id.backgroundImageView) as ImageView
        BT_viewUtilities.updateBackgroundImageForScreen(backgroundImageView, tmpItem)

        // setup transition delay if not -1 (-1 means tap to start animation) 
        if (startTransitionAfterSeconds > -1) {
            launch {
                delay(1000L * (startTransitionAfterSeconds + 1))
                animateSplashScreen()
            }
        }
    }

    // handle touch event..
    override fun onTouch(arg0: View, arg1: MotionEvent): Boolean {
        if (startTransitionAfterSeconds < 1) {
            animateSplashScreen()
        }
        return false
    }

    // startEffect...
    private fun startEffect(effectType: String, imageView: ImageView?) {
        BT_debugger.showIt("$fragmentName:startEffect. \"$effectName\"")

        // Effect: "movingFade" is an image overlay that animates left to right...
        if (effectType.equals("fallDown", ignoreCase = true) && imageView != null) {

            // animation comes from .xml resource...
            val anim = AnimationUtils.loadAnimation(activity, R.anim.bt_screen_splash_effect_falldown)
            imageView.startAnimation(anim)
        } // movingFade...
    }

    private fun onAnimationFinished() {
        BT_debugger.showIt("$fragmentName:animationFinished")

        // start the BT_activity_host activity...
        val i = Intent(activity, if (buzztouch_appDelegate.rootApp.tabs.size > 0) BT_activity_host_tabs::class.java else BT_activity_host::class.java)
        startActivity(i)

        // finish this activity (remove it from the back stack)...
        activity?.finish()

        // fade in next screen...
        activity?.overridePendingTransition(R.anim.bt_fadein, R.anim.bt_fadeout)
    }

    // end delay stuff
    // ///////////////////////////////////////////////////////////////////
    // handles animation...
    private fun animateSplashScreen() {
        BtLog("$fragmentName:animateSplashScreen")

        /*
			This fragment should be running in BT_activity_splash. The layout file for this
			Android Activity is bt_activity_splash.xml. Get a reference to the "containerView"
			(a FrameLayout) in the layout file so we can animate it...	
		*/try {

            // get a reference to the host activities "containerView" This is a FrameLayout ...
            val containerView = this.activity?.findViewById<View>(R.id.containerView) as FrameLayout

            // animation comes from .xml resource...
            val anim = AnimationUtils.loadAnimation(activity, R.anim.bt_fadeout)
            anim.duration = (transitionDurationSeconds * 1000).toLong()
            containerView.startAnimation(anim)
            containerView.visibility = View.GONE

            // delay the transition...
            launch {
                delay(1000L * transitionDurationSeconds)
                onAnimationFinished()
            }
        } catch (e: NullPointerException) {
            BtLog("$fragmentName:animateSplashScreen EXCEPTION: $e")
        }
    }
}
