/*  File Version: 3.0
 *	Copyright, David Book, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.utils

import android.content.Context
import android.graphics.Bitmap
import android.graphics.Bitmap.CompressFormat
import android.graphics.BitmapFactory
import android.graphics.drawable.Drawable
import androidx.core.content.res.ResourcesCompat
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.extensions.*
import java.io.*
import java.util.*
import kotlin.collections.ArrayList

object BT_fileManagerKt {
    var TAG = "BT_fileManager"

    // not all documents are cachable...
    fun canCacheDocumentType(theFileName: String): Boolean {
        BtLog("$TAG:canCacheDocument \"$theFileName\"")
        val cacheableTypes = listOf(".htm", ".txt", ".pdf", ".xls", ".ppt", ".doc")

        return cacheableTypes.any { theFileName.contains(it) }
    }

    // does file exist in projects assets directory? assetsDirectory will be "BT_Images" or "BT_Audio", etc...
    fun doesProjectAssetExist(assetsDirectory: String, fileName: String): Boolean {
        val am = buzztouch_appDelegate.getApplication().assets
        return (
            try {
                am.list(assetsDirectory)?.any {
                    it.equals(fileName, ignoreCase = true)
                } ?: false
            } catch (e: java.lang.Exception) {
                BtLog("$TAG:doesProjectAssetExist EXCEPTION $e")
                e.printStackTrace()
                false
            }
            ).also { if (!it) BtLog("$TAG :doesProjectAssetExist: NO assets/$assetsDirectory/fileName") }
    }

    // copies a file from assets to the app's cache..
    fun copyAssetToCache(assetsDirectory: String, fileName: String): Boolean {
        BtLog("$TAG:copyAssetToCache: $fileName")

        return buzztouch_appDelegate.getAssetManager().readText("$assetsDirectory/$fileName")?.let {
            buzztouch_appDelegate.getApplication().writeTextToCache(fileName, it)
        } ?: false
    }

    // copies a file from the app's cache to the SDCARD..
    fun copyFileFromCacheToSDCard(fileName: String): Boolean {
        BtLog("$TAG:copyFileFromCacheToSDCard: $fileName")
        // return...
        return try {
            // copy from assets to internal cache...
            val res = File(fileName).copyTo(File(buzztouch_appDelegate.getApplication().externalCacheDir, fileName))
            res.exists()
        } catch (e: Exception) {
            BtLog("$TAG:copyFileFromCacheToSDCard :EXCEPTION $e")
            e.printStackTrace()
            false
        }
    }

    // returns a list of all the assets in a folder. Much faster to search an array than the file system!
    fun getAssetListFromDirectory(assetsDirectory: String): ArrayList<String> {
        val am = buzztouch_appDelegate.getAssetManager()
        return try {
            ArrayList(am.list(assetsDirectory)?.filter { it.length > 3 } ?: listOf())
        } catch (e: Exception) {
            BtLog("$TAG:getAssetsInDirectory: EXCEPTION loading assets list from $assetsDirectory")
            e.printStackTrace()
            ArrayList()
        }
    }

    // does file exist in applications download cache?
    fun doesCachedFileExist(fileName: String): Boolean = try {
        buzztouch_appDelegate.getApplication().getFileStreamPath(fileName).exists()
    } catch (e: Exception) {
        e.printStackTrace()
        false
    }

    // returns resource id from projects bundle
    fun getResourceIdFromBundle(resourcesDirectory: String?, fileName: String): Int {
        // remove the file extension form the file name (not used when working with resources)
        val useResName: String = BT_stringsKt.removeExtension(fileName)
        return try {
            buzztouch_appDelegate.getApplication().resources.getIdentifier(useResName, resourcesDirectory, buzztouch_appDelegate.getApplication().packageName)
        } catch (e: Exception) {
            BtLog("$TAG: An exception occurred trying to find the resource id for: $fileName")
            0
        }
    }

    fun getResourceId(name: String, type: String): Int =
        buzztouch_appDelegate.getApplication().resources.getIdentifier(name, type, buzztouch_appDelegate.getApplication().packageName)

    // returns drawable by name
    fun getDrawableByName(fileName: String): Drawable? {
        return try { ResourcesCompat.getDrawable(appContext.resources, getResourceId(fileName, "drawable"), null) } catch (e: Exception) {
            BtLog("$TAG:getDrawableFromResourcesByName An exception occurred trying to get the drawable named: $fileName")
            null
        }
    }

    // get image (Bitmap) from URL...
    suspend fun getImageFromURL(theUrl: String): Bitmap? = appContext.downloadImage(theUrl)?.bitmap

    // get image (drawable) from cache...
    fun getDrawableFromCache(fileName: String): Drawable? {
        return try {
            appContext.openFileInput(fileName)?.let {
                val d = Drawable.createFromStream(it, null)
                it.close()
                d
            } ?: null.also { BtLog("$TAG:getDrawableFromCache Could not find \"$fileName\" in application's cache directory") }
        } catch (c: Exception) {
            BtLog("$TAG:getDrawableFromCache An exception occurred trying to get an image from the cache: $c")
            null
        }
    }

    // get image (Bitmap) from cache...
    fun getBitmapFromCache(fileName: String): Bitmap? {
        return try {
            appContext.openFileInput(fileName)?.let {
                val b = BitmapFactory.decodeStream(it)
                it.close()
                b
            } ?: null.also { BtLog("\"$TAG:getBitmapFromCache Could not find \\\"$fileName\\\" in application's cache directory") }
        } catch (e: Exception) {
            BtLog("$TAG:getBitmapFromCache An exception occurred trying to get an image from the cache: $fileName")
            null
        }
    }

    // save image to cache
    fun saveImageToCache(theImage: Bitmap?, saveAsFileName: String) {
        BtLog("$TAG: saveImageToCache: $saveAsFileName")
        if (saveAsFileName.length <= 3) return BtLog("$TAG :saveImageToCache() invalid saveAsFileName: $saveAsFileName")
        if (theImage == null) return BtLog("$TAG :saveImageToCache() image is null")
        try {
            val fos = appContext.openFileOutput(saveAsFileName, Context.MODE_PRIVATE)
            theImage.compress(CompressFormat.JPEG, 100, fos)
            fos.flush()
            fos.close()
        } catch (e: Exception) {
            BtLog("$TAG: An exception occurred trying to save an image to the cache? $e")
        }
    }

    // save file to cache
    fun saveTextFileToCache(theData: String?, saveAsFileName: String) {
        BtLog("$TAG: saveTextFileToCache: $saveAsFileName")
        try {
            if (saveAsFileName.length > 3) {
                val out = OutputStreamWriter(appContext.openFileOutput(saveAsFileName, Context.MODE_PRIVATE))
                out.write(theData)
                out.close()
            }
        } catch (e: Exception) {
            BtLog("$TAG: An exception occurred trying to save a text file to the cache? $e")
        }
    }

    // get text file contents from project assets...
    fun readTextFileFromAssets(assetsDirectory: String, fileName: String): String {
        BtLog("$TAG: readTextFileFromAssets: \"$assetsDirectory/$fileName\"")
        val am = appContext.resources.assets
        return try {
            val inSt = if (assetsDirectory.isEmpty()) am.open(fileName) else am.open("$assetsDirectory/$fileName")
            (inSt.readText() ?: "").also { inSt.close() }
        } catch (e: Exception) {
            BtLog("$TAG:readTextFileFromAssets: EXCEPTION reading text file from assets/$assetsDirectory/$fileName Error: $e")
            ""
        }
    }

    // get text file contents from cache...
    fun readTextFileFromCache(fileName: String): String {
        BtLog("$TAG: readTextFileFromCache: \"$fileName\"")
        if (fileName.isEmpty()) return "".also { BT_debugger.logError(buzztouch_appDelegate.getContext(), "Error: readTextFileFromCache where fileName is missing") }
        return try {
            appContext.openFileInput(fileName).let { fIs ->
                fIs.readText().also { fIs.close() } ?: ""
            }
        } catch (e: Exception) {
            BtLog("$TAG: ERROR 2. An exception occurred trying to read $fileName from the cache.")
            e.printStackTrace()
            ""
        }
    } // return

    // get size of cached data directory..
    val cachedDataSize: Int
        get() {
            BtLog("$TAG:getCachedDataSize")
            return try {
                val appDataDirectory = appContext.filesDir.absolutePath
                if (appDataDirectory.length > 3) {
                    try {
                        File(appDataDirectory).listFiles()?.sumBy { it.length().toInt() } ?: 0
                    } catch (e: Exception) {
                        BtLog("$TAG:deleteAllCachedData Exception: $e")
                        0
                    }
                } else 0
            } catch (e: Exception) {
                BtLog("$TAG: EXCEPTION $e")
                0
            }
        }

    /**
     * Loads a raw resource and returns its content as a String.
     *
     * @throws IOException If any error was encountered, such as an incorrect resource ID, or
     * inaccessible file.
     */
    @Throws(IOException::class)
    fun parseResource(context: Context, resource: Int): String? {
        return try { appContext.resources.openRawResource(resource).readText() } catch (e: Exception) {
            e.printStackTrace()
            null
        }
    }

    // deletes one cached file...
    fun deleteFile(fileName: String) {
        BtLog("$TAG:deleteFile $fileName")
        val file = appContext.getFileStreamPath(fileName)
        if (file.exists()) {
            BtLog("$TAG:deleteFile deleting $fileName")
            file.delete()
        } else {
            BtLog("$TAG:deleteFile $fileName does not exist")
        }
    }

    // deletes all cached files...
    fun deleteAllCachedData(skipFile: String?) {
        BtLog("$TAG: deleting application cache")
        val appDataDirectory = appContext.filesDir.absolutePath
        BtLog("$TAG: deleting application cache in: $appDataDirectory")
        if (appDataDirectory.length > 3) {
            try {
                File(appDataDirectory)
                    .listFiles()
                    ?.filterNot { it.name.equals(skipFile, ignoreCase = true) }
                    ?.filterNot { it.name.startsWith("persist_") }
                    ?.forEach {
                        BtLog("$TAG: deleting: ${it.name}")
                        it.delete()
                    }
            } catch (e: Exception) {
                BtLog("$TAG:deleteAllCachedData Excpetion: $e")
            }
        }
    }
}
