/*  File Version: 3.0
 *	Copyright, David Book, buzztouch.com
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are 
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the 
 *	following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list 
 *	of conditions and the following disclaimer in the documentation and/or other materials 
 *	provided with the distribution. 
 *
 *	Neither the name of David Book, or buzztouch.com nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software without specific 
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 *	OF SUCH DAMAGE. 
 */
package com.buzztouch.utils

import android.net.Uri
import com.buzztouch.buzztouch_appDelegate
import com.buzztouch.data_models.BT_item
import com.buzztouch.extensions.BtLog
import com.buzztouch.extensions.getStringOrNull
import com.buzztouch.extensions.second
import org.json.JSONException
import org.json.JSONObject
import java.io.File
import java.net.URLEncoder
import java.util.*
import kotlin.math.ln
import kotlin.math.pow

object BT_stringsKt {
    private const val TAG = "BT_strings"

    // removes file extension from a file name
    fun removeExtension(s: String): String {
        return s.substringAfterLast(File.separator).substringBeforeLast(".")
    }

    // get property value from JSON string...
    fun getJsonPropertyValue(theJSONObject: JSONObject?, nameOfProperty: String?, defaultValue: String?): String? {
        return theJSONObject?.getStringOrNull(nameOfProperty ?: "") ?: defaultValue
    }

    // returns a style value for a screen setting from theme or screen data...
    fun getStyleValueForScreen(theParentScreenData: BT_item, nameOfProperty: String, defaultValue: String): String {
        // global theme is used if screen does not have a setting
        val theThemeData: JSONObject = buzztouch_appDelegate.rootApp.rootTheme.jsonObject
        return theParentScreenData.jsonObject.getStringOrNull(nameOfProperty)
            ?: theThemeData.getStringOrNull(nameOfProperty)
            ?: defaultValue
    }

    /**
     * Converts a Map object to String in format of www-form-encoded
     * @param map -- Map object to convert
     * @return The encoded String to return
     */
    fun convertMapToEncodedString(map: Map<String?, String?>): String =
        map.entries.joinToString("&") { "${it.key}=${it.value}" }

    // gets a "saveAsFileName" from a URL
    fun getSaveAsFileNameFromURL(theURL: String): String {
        // if we have a 'saveAsFileName' in the querystring, return that..else..return file name
        val uri = Uri.parse(theURL)
        val default = File(uri.toString()).name
        val ret = try {
            uri.getQueryParameter("saveAsFileName") ?: default
        } catch (n: Exception) {
            default
        }
        BtLog("$TAG:getSaveAsFileNameFromURL from: \"$theURL\" file name: \"$ret\"")
        return ret
    }

    // gets a value of a preference saved on the device...
    @Deprecated("Use BT_preferences class", ReplaceWith("BT_preferences.getPrefString(nameOfPreference)"))
    fun getPrefString(nameOfPreference: String?): String {
        return BT_preferences.getPrefString(nameOfPreference)
    }

    // sets a value of a preference to the device...
    @Deprecated("Use BT_preferences class", ReplaceWith("BT_preferences.setPrefString(nameOfPreference, valueOfPreference)"))
    fun setPrefString(nameOfPreference: String?, valueOfPreference: String?) {
        BT_preferences.setPrefString(nameOfPreference, valueOfPreference)
    }

    // gets a value of a preference saved on the device...
    @Deprecated("Use BT_preferences class", ReplaceWith("BT_preferences.getPrefInteger(nameOfPreference)"))
    fun getPrefInteger(nameOfPreference: String?): Int {
        return BT_preferences.getPrefInteger(nameOfPreference)
    }

    // sets a value of a preference to the device...
    @Deprecated("Use BT_preferences class", ReplaceWith("BT_preferences.setPrefInteger(nameOfPreference, valueOfPreference)"))
    fun setPrefInteger(nameOfPreference: String?, valueOfPreference: Int?) {
        BT_preferences.setPrefInteger(nameOfPreference, valueOfPreference)
    }

    // delete local prefs
    @Deprecated("Use BT_preferences class", ReplaceWith("BT_preferences.deleteAllLocalPrefs()"))
    fun deleteAllLocalPrefs() {
        BT_preferences.deleteAllLocalPrefs()
    }

    // merges variables in string (usually a URL)
    fun mergeBTVariablesInString(theString: String): String {
        return buzztouch_appDelegate.rootApp?.let { app ->
            theString.apply {
                replace("[buzztouchAppId]", app.buzztouchAppId)
                replace("[buzztouchAPIKey]", buzztouch_appDelegate.rootApp.buzztouchAPIKey)
                app.currentScreenData?.let { replace("[screenId]", it.itemId) }
                app.rootUser?.let { user ->
                    replace("[userId]", user.userId)
                    replace("[userEmail]", user.userEmail)
                    replace("[userLogInId]", user.userLogInId)
                    replace("[userLogInPassword]", user.userLogInPassword)
                }
                app.rootDevice?.let { device ->
                    replace("[deviceId]", device.deviceId)
                    replace("[deviceLatitude]", device.deviceLatitude)
                    replace("[deviceLongitude]", device.deviceLongitude)
                    replace("[deviceModel]", device.deviceModel)
                }
            }
        } ?: theString
    }

    // formats bytes to human readable string
    fun formatBytes(bytes: Long, siPrefix: Boolean = true): String {
        BtLog("$TAG:formatBytes Formatting: $bytes")
        val unit = if (siPrefix) 1000 else 1024
        if (bytes < unit) return "$bytes B"
        val exp = (ln(bytes.toDouble()) / ln(unit.toDouble())).toInt()
        val pre = (if (siPrefix) "kMGTPE" else "KMGTPE")[exp - 1].toString() + if (siPrefix) "" else "i"
        return String.format("%.1f %sB", bytes / unit.toDouble().pow(exp.toDouble()), pre)
    }

    // returns a JSON object from a string...
    fun getJsonObjectFromString(theJsonString: String?): JSONObject? {
        // BtLog(objectName + ":getJsonObjectFromString String: " + theJsonString);
        theJsonString ?: return null
        return try {
            JSONObject(theJsonString)
        } catch (e: JSONException) {
            BtLog("$TAG:getJsonObjectFromString EXCEPTION: $e")
            e.printStackTrace()
            null
        }
    }

    // returns a URL encoded string...
    fun urlEncode(theString: String?): String {
        // BtLog(objectName + ":urlEncode: " + theString);
        theString ?: return ""
        return try { URLEncoder.encode(theString, "utf-8").replace("+", "%20") } catch (e: Exception) {
            e.printStackTrace()
            ""
        }
    }

    // returns a URL encoded string from a complete URL...
    fun encodeCompleteURL(theURL: String?): String {
        // pull the theURL apart and URLEncode the params...
        val uri = Uri.parse(theURL)
        val hostPart = uri.scheme + "://" + uri.host + uri.path
        val qryPart = uri.query

        // clean up query query component...
        val params: List<Pair<String, String>> = qryPart?.split("&")?.map { param ->
            param.split("=").let { parts ->
                when (parts.size) {
                    2 -> Pair(parts.first(), parts.second())
                    1 -> Pair(parts.first(), "")
                    else -> Pair("", "")
                }
            }
        } ?: listOf()
        val qryString = params.joinToString("&") { "${it.first}=${it.second}" }

        return "$hostPart?$qryString"
    }

    // creates a random guid
    fun createGuid(): String = UUID.randomUUID().toString()
}
